/********************************************************************************
 * Copyright (C) 2020 Genuitec, LLC. All rights reserved.
 *
 * Usage of this software is governed by the CodeTogether End User License Agreement.
 *
 * SPDX-License-Identifier: SEE LICENSE IN LICENSE
 ********************************************************************************/

import { inject, injectable } from 'inversify';
import { KeybindingRegistry, KeybindingContribution } from '../keybinding';
import { PresenceMemento } from '../codetogether-startup/presence-memento';
import { Keybinding } from '../../common/keybinding';
import { IDEKeybindings, KeybindingsStyle } from './ct-ide-keybindings';
import { IDE } from '../codetogether-startup/session';

export const KEYBINDINGS_KEY = 'keybindings_style';

export function getKeybindingsStyleFromIDE(ide: IDE): KeybindingsStyle {
    switch (ide) {
        case IDE.Eclipse:
            return KeybindingsStyle.Eclipse;
        case IDE.IntelliJ:
            return KeybindingsStyle.IntelliJ;
        default:
            return KeybindingsStyle.VSCode;
    }
}

@injectable()
export class KeybindingsService implements KeybindingContribution {
    constructor(
        @inject(PresenceMemento) protected readonly presenceMemento: PresenceMemento,
        @inject(KeybindingRegistry) protected readonly keybindingRegistry: KeybindingRegistry
    ) { }

    async registerKeybindings(keybindingsRegistry: KeybindingRegistry): Promise<void> {
        const keybindingsStyle = await this.getKeybindingsStyle();
        if (keybindingsStyle) {
            this.updateKeybindings(keybindingsStyle, keybindingsRegistry);
        }
    }

    private updateKeybindings(style: KeybindingsStyle, keybindingsRegistry: KeybindingRegistry): void {
        const keybindings = this.getKeybindingsFromStyle(style);
        keybindings.reduce(
            (groupedKeybinding, keybinding) => {
                let processedKeybindings = groupedKeybinding.get(keybinding.keybinding);
                if (!processedKeybindings) {
                    processedKeybindings = [];
                    groupedKeybinding.set(keybinding.keybinding, processedKeybindings);
                }
                processedKeybindings.push(keybinding);
                return groupedKeybinding;
            },
            new Map<string, Keybinding[]>()
            ).forEach((keybindingEntries, keybinding, groupedKeybindings) => {
            // Remove preexisting keybinding
            keybindingsRegistry.unregisterKeybinding(keybinding);
            keybindingEntries.forEach(keybindingEntry => {
                /**
                 * Remove action keybinding if keybinding doesn't match ours
                 */
                keybindingsRegistry.getKeybindingsForCommand(keybindingEntry.command)
                    .forEach(keybindingForCommand => {
                        const ourKeybinding = groupedKeybindings.get(keybindingForCommand.keybinding);
                        if (ourKeybinding) {

                            // Make sure command and keybinding are the same
                            const matchesOurKeybinding = ourKeybinding.some(k => k.command === keybindingForCommand.command);
                            if (!matchesOurKeybinding) {
                                keybindingsRegistry.unregisterKeybinding(keybindingForCommand);
                            }
                        } else {
                            keybindingsRegistry.unregisterKeybinding(keybindingForCommand);
                        }
                    });
            });
            keybindingEntries.forEach(keybindingEntry => keybindingsRegistry.registerKeybinding(keybindingEntry));
        });
    }

    setKeybindingsStyle(style: KeybindingsStyle): Promise<void> {
        if (this.keybindingRegistry) {
            this.updateKeybindings(style, this.keybindingRegistry);
        }
        return this.presenceMemento.update(KEYBINDINGS_KEY, style);
    }

    private getKeybindingsFromStyle(style: KeybindingsStyle): Keybinding[] {
        switch (style) {
            case KeybindingsStyle.Eclipse:
                return IDEKeybindings.Eclipse;
            case KeybindingsStyle.IntelliJ:
                return IDEKeybindings.IntelliJ;
            default:
                return IDEKeybindings.VSCode;
        }
    }

    getKeybindingsStyle(): Promise<KeybindingsStyle | undefined> {
        return Promise.resolve(this.presenceMemento.get(KEYBINDINGS_KEY));
    }
}
