/********************************************************************************
 * Copyright (C) 2020 Genuitec, LLC. All rights reserved.
 *
 * Usage of this software is governed by the CodeTogether End User License Agreement.
 *
 * SPDX-License-Identifier: SEE LICENSE IN LICENSE
 ********************************************************************************/

import { injectable, inject } from 'inversify';
import axios from 'axios';
import { UUID } from '@phosphor/coreutils';
import { WebSocketChannel, ISessionProvider } from '../common/messaging/web-socket-channel';
import { Emitter, Event } from '../common/event';
import { Endpoint } from './endpoint';

const CT_BASE_URL = 'https://apps.cloud.genuitec.com';

export const WebSocketUrlProvider = Symbol('WebSocketUrlProvider');
export const SessionProviderService = Symbol('SessionProviderService');

@injectable()
export class CTWebSocketUrlProvider {

    private valid = false;
    private wsURL: string | undefined;

    @inject(SessionProviderService) protected sessionProvider: SessionProvider;

    public getUrl(): string | undefined {
        return this.wsURL;
    }
    public getWebsocketUrl(): Promise<string> {
        if (this.wsURL) {
            return Promise.resolve(this.wsURL);
        }
        return new Promise(resolve => {
            const currentLocation = window.location;
            const requestHeaders = {
                'ctorigin': currentLocation.host.indexOf(':') >= 0 ? currentLocation.host.substring(0, currentLocation.host.indexOf(':')) : currentLocation.host
            };
            // tslint:disable-next-line: no-any
            axios.get('/ct-url').then((_resp: any) => {
                let baseUrl = 'https://apps.cloud.genuitec.com';
                let devUrl = false;
                if (_resp.status === 200 && _resp.data && _resp.data.direct) {
                    this.valid = true;
                    this.wsURL = this.createWebSocketUrl(WebSocketChannel.wsPath);
                    resolve(this.wsURL);
                    return;
                }
                if (_resp.status === 200 && _resp.data && _resp.data.url) {
                    baseUrl = _resp.data.url;
                }
                if (baseUrl !== CT_BASE_URL) {
                    devUrl = true;
                }
                // v1 infor does not need jwt token
                axios.post(baseUrl + '/ct/session/v1/info', {
                    uuid: this.sessionProvider.session
                    // tslint:disable-next-line: no-any
                }, { headers: requestHeaders }).then((resp: any) => {
                    if (resp.status === 200 && resp.data) {
                        const host = resp.data.host;
                        const port = resp.data.port;
                        // to be able to do a cors websocket connection we first need to do an http to that host
                        // https://stackoverflow.com/questions/22644392/chrome-websockets-cors-policy
                        const target = (devUrl && (host === '127.0.0.1' || host === 'localhost'))
                            ? 'http://' + host + ':' + port + '/ct-setup' : 'https://' + host + ':' + port + '/ct-setup';
                        axios.get(target).then(_ => {
                            const schema = (devUrl && (host === '127.0.0.1' || host === 'localhost')) ? 'ws' : 'wss';
                            this.wsURL = schema + '://' + host + ':' + port + '' + WebSocketChannel.wsPath;
                            resolve(this.wsURL);
                            this.valid = true;
                        }).catch(_err => {
                            const schema = (devUrl && (host === '127.0.0.1' || host === 'localhost')) ? 'ws' : 'wss';
                            this.wsURL = schema + '://' + host + ':' + port + '' + WebSocketChannel.wsPath;
                            this.valid = true;
                            resolve(this.wsURL);
                        });
                        return;
                    }
                    this.valid = false;
                    resolve('');
                    // tslint:disable-next-line: no-any
                }).catch((err: any) => {
                    this.valid = false;
                    resolve('');
                });
            });
        });
    }

    validateLocation(): void {
        if (!this.wsURL && window.location.hostname.indexOf('127.0.0.1') === -1 && window.location.hostname.indexOf('localhost') === -1 && !window.location.hostname.startsWith('10.')) {
            setTimeout(() => {
                window.location.assign('https://codetogether.com');
            }, 150);
        }
    }

    private createWebSocketUrl(path: string): string {
        const endpoint = new Endpoint({ path });
        return endpoint.getWebSocketUrl().toString();
    }

    public isValid(): boolean {
        return this.valid;
    }
}

export const onBrowserSessionEmitter: Emitter<string> = new Emitter();
export const onBrowserSessionEndEmitter: Emitter<void> = new Emitter();

@injectable()
export class SessionProvider implements ISessionProvider {
    session: string;
    participantID: string;
    readonly onBrowserSessionSetup: Event<string>;
    readonly onBrowserSessionEnd: Event<void>;
    constructor() {
        const currentLocation = window.location;
        let hash = currentLocation.hash.substr(1);
        if (hash && hash.length && hash.indexOf('/') >= 0) {
            hash = hash.substr(1);
            if (hash && hash.length && hash.indexOf('/') >= 0) {
                hash = hash.substr(0, hash.indexOf('/'));
            }
        }
        this.session = hash;
        this.participantID = UUID.uuid4();
        this.onBrowserSessionSetup = onBrowserSessionEmitter.event;
        this.onBrowserSessionEnd = onBrowserSessionEndEmitter.event;
    }
}
