/********************************************************************************
 * Copyright (C) 2017 TypeFox and others.
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * This Source Code may also be made available under the following Secondary
 * Licenses when the conditions for such availability set forth in the Eclipse
 * Public License v. 2.0 are satisfied: GNU General Public License, version 2
 * with the GNU Classpath Exception which is available at
 * https://www.gnu.org/software/classpath/license.html.
 *
 * SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
 ********************************************************************************/

import { inject, injectable, named, interfaces } from 'inversify';
import { ContributionProvider, CommandRegistry, MenuModelRegistry, ILogger, isOSX } from '../common';
import { MaybePromise } from '../common/types';
import { KeybindingRegistry } from './keybinding';
import { Widget } from './widgets';
import { ApplicationShell } from './shell/application-shell';
import { ShellLayoutRestorer, ApplicationShellLayoutMigrationError } from './shell/shell-layout-restorer';
import { FrontendApplicationStateService } from './frontend-application-state';
import { preventNavigation, parseCssTime } from './browser';
import { CorePreferences } from './core-preferences';
import { CTWebSocketUrlProvider, SessionProvider, onBrowserSessionEndEmitter, WebSocketUrlProvider, SessionProviderService } from './ct-websocket-url-provider';
import { WebSocketConnectionProvider } from './messaging';
import { sessionServicePath, ISessionService, SessionDetails, EndToEndServiceImpl, EndToEndService, IDE } from './codetogether-startup/session';
import { renderCodeTogetherStartup } from './codetogether-startup/codetogether-startup';
import { PresenceMemento } from './codetogether-startup/presence-memento';
import { KeybindingsService, getKeybindingsStyleFromIDE } from './ct-keybindings/ct-keybindings-service';
import { KeybindingsStyle } from './ct-keybindings/ct-ide-keybindings';
import { ThemeIDE, ThemeIDEService, ideToDefaultThemeIDE } from './ct-theme-ide/ct-theme-ide-service';
import { ThemeService } from './theming';

/**
 * Clients can implement to get a callback for contributing widgets to a shell on start.
 */
export const FrontendApplicationContribution = Symbol('FrontendApplicationContribution');
export interface FrontendApplicationContribution {

    /**
     * Called on application startup before configure is called.
     */
    initialize?(): void;

    /**
     * Called before commands, key bindings and menus are initialized.
     * Should return a promise if it runs asynchronously.
     */
    configure?(app: FrontendApplication): MaybePromise<void>;

    /**
     * Called when the application is started. The application shell is not attached yet when this method runs.
     * Should return a promise if it runs asynchronously.
     */
    onStart?(app: FrontendApplication): MaybePromise<void>;

    /**
     * Called on `beforeunload` event, right before the window closes.
     * Return `true` in order to prevent exit.
     * Note: No async code allowed, this function has to run on one tick.
     */
    onWillStop?(app: FrontendApplication): boolean | void;

    /**
     * Called when an application is stopped or unloaded.
     *
     * Note that this is implemented using `window.beforeunload` which doesn't allow any asynchronous code anymore.
     * I.e. this is the last tick.
     */
    onStop?(app: FrontendApplication): void;

    /**
     * Called after the application shell has been attached in case there is no previous workbench layout state.
     * Should return a promise if it runs asynchronously.
     */
    initializeLayout?(app: FrontendApplication): MaybePromise<void>;

    /**
     * An event is emitted when a layout is initialized, but before the shell is attached.
     */
    onDidInitializeLayout?(app: FrontendApplication): MaybePromise<void>;
}

/**
 * Default frontend contribution that can be extended by clients if they do not want to implement any of the
 * methods from the interface but still want to contribute to the frontend application.
 */
@injectable()
export abstract class DefaultFrontendApplicationContribution implements FrontendApplicationContribution {

    initialize(): void {
        // NOOP
    }

}

@injectable()
export class FrontendApplication {

    private app: FrontendApplicationDelegate;

    public container: interfaces.Container;
    private session: string | undefined;
    private signature: string | undefined;
    private presenceMemento: PresenceMemento;
    public keybindingsService: KeybindingsService;

    private startupElem: HTMLElement | undefined;
    constructor(
        @inject(WebSocketUrlProvider) protected readonly webSocketUrlProvider: CTWebSocketUrlProvider,
        @inject(SessionProviderService) protected sessionProvider: SessionProvider,
        @inject(EndToEndService) public endToEndService: EndToEndServiceImpl,
        @inject(ThemeIDEService) public themeIDEService: ThemeIDEService,
        @inject(ThemeService) public themeService: ThemeService
    ) {
        this.presenceMemento = new PresenceMemento();
    }

    get shell(): ApplicationShell {
        return this.app.shell;
    }

    async start(): Promise<void> {
        const parts = window.location.hash.split('/');
        if (parts.length >= 3) {
            this.session = parts[1];
            this.signature = parts[2];
        }

        return this.webSocketUrlProvider.getWebsocketUrl().then(_url => {
            const sessionService = WebSocketConnectionProvider.createProxy<ISessionService>(this.container, sessionServicePath);
            this.app = this.container.get(FrontendApplicationDelegate);
            return new Promise(resolve => {
                sessionService.getSessionDetails(this.getSession()).then(details => {
                    this.revealShell().then(_ => this.getHost().then(host =>
                        this.startCodeTogether(host, sessionService, details, this.presenceMemento).then(() => resolve())));
                });
                window.addEventListener('hashchange', async () => {
                    if (window.location.hash === '') {
                        // Trigger to disconnect websocket
                        onBrowserSessionEndEmitter.fire();
                    } else {
                        window.location.reload();
                    }

                });
            });
        });
    }

    private startCodeTogether(
        body: HTMLElement, sessionService: ISessionService, sessionDetails: SessionDetails,
        presenceMemento: PresenceMemento
    ): Promise<void> {
        const ide: IDE = sessionDetails.ide as IDE;
        let passwordPromise: Promise<void>;
        let keybindingsStyle: KeybindingsStyle;
        let themeIDE: ThemeIDE;
        let showOverlay: boolean;
        if (this.signature) {
            if (this.session) {
                this.endToEndService.session = this.session;
            }
            passwordPromise = this.endToEndService.setPassword(this.signature);
        } else {
            passwordPromise = Promise.resolve();
        }

        return passwordPromise.then(() => {
            this.keybindingsService = this.container.get(KeybindingsService);
            this.keybindingsService.getKeybindingsStyle().then(currentKeybindingsStyle => {
                if (currentKeybindingsStyle) {
                    keybindingsStyle = currentKeybindingsStyle;
                } else {
                    keybindingsStyle = getKeybindingsStyleFromIDE(ide);
                    this.keybindingsService.setKeybindingsStyle(keybindingsStyle);
                }
            });
        }).then(() =>
            this.themeIDEService.getThemeIDE().then(currentThemeIDE => {
                showOverlay = !currentThemeIDE;
                if (currentThemeIDE) {
                    themeIDE = currentThemeIDE;
                } else {
                    themeIDE = ideToDefaultThemeIDE(ide);
                    this.themeIDEService.setThemeIDE(themeIDE, true);
                }
            })).then(() => {
                const ctEntry = document.createElement('div');
                ctEntry.setAttribute('id', 'codetogether');
                ctEntry.classList.add('ct-startup-out-box');
                body.appendChild(ctEntry);
                renderCodeTogetherStartup(ctEntry, this, sessionService, sessionDetails, presenceMemento, keybindingsStyle, themeIDE, showOverlay);
                return Promise.resolve();
            });
    }

    public startTheia(): Promise<void> {
        return this.revealPreload().then(_ => this.app.start());
    }

    public getSession(): string {
        return this.session || '';
    }

    public getPassword(): string {
        return this.signature || '';
    }

    private revealPreload(): Promise<void> {
        return this.getHost().then(host => {
            const ctElement = document.getElementById('codetogether');
            if (this.startupElem) {
                this.startupElem.classList.remove('ct-preload-hidden');
                host.appendChild(this.startupElem);
            }
            if (ctElement) {
                host.removeChild(ctElement);
            }
        });
    }

    public revealShell(): Promise<void> {
        return this.getHost().then(host => {
            this.startupElem = this.app.getStartupIndicator(host);
            if (this.startupElem) {
                this.startupElem.classList.add('ct-preload-hidden');
            }
        });
    }

    private getHost(): Promise<HTMLElement> {
        if (document.body) {
            return Promise.resolve(document.body);
        }
        return new Promise<HTMLElement>(resolve =>
            window.addEventListener('load', () => resolve(document.body), { once: true })
        );
    }
}

@injectable()
export class FrontendApplicationDelegate {

    @inject(CorePreferences)
    protected readonly corePreferences: CorePreferences;

    constructor(
        @inject(CommandRegistry) protected readonly commands: CommandRegistry,
        @inject(MenuModelRegistry) protected readonly menus: MenuModelRegistry,
        @inject(KeybindingRegistry) protected readonly keybindings: KeybindingRegistry,
        @inject(ILogger) protected readonly logger: ILogger,
        @inject(ShellLayoutRestorer) protected readonly layoutRestorer: ShellLayoutRestorer,
        @inject(ContributionProvider) @named(FrontendApplicationContribution)
        protected readonly contributions: ContributionProvider<FrontendApplicationContribution>,
        @inject(ApplicationShell) protected readonly _shell: ApplicationShell,
        @inject(FrontendApplicationStateService) protected readonly stateService: FrontendApplicationStateService,
        @inject(ThemeService) public themeService: ThemeService
    ) { }

    get shell(): ApplicationShell {
        return this._shell;
    }

    /**
     * Start the frontend application.
     *
     * Start up consists of the following steps:
     * - start frontend contributions
     * - attach the application shell to the host element
     * - initialize the application shell layout
     * - reveal the application shell if it was hidden by a startup indicator
     */
    async start(): Promise<void> {
        await this.startContributions();
        this.stateService.state = 'started_contributions';

        const host = await this.getHost();
        this.attachShell(host);
        await new Promise(resolve => requestAnimationFrame(() => resolve()));
        this.stateService.state = 'attached_shell';

        await this.initializeLayout();
        this.stateService.state = 'initialized_layout';
        await this.fireOnDidInitializeLayout();

        await this.revealShell(host);
        this.registerEventListeners();
        this.stateService.state = 'ready';
    }

    /**
     * Return a promise to the host element to which the application shell is attached.
     */
    protected getHost(): Promise<HTMLElement> {
        if (document.body) {
            return Promise.resolve(document.body);
        }
        return new Promise<HTMLElement>(resolve =>
            window.addEventListener('load', () => resolve(document.body), { once: true })
        );
    }

    /**
     * Return an HTML element that indicates the startup phase, e.g. with an animation or a splash screen.
     */
    public getStartupIndicator(host: HTMLElement): HTMLElement | undefined {
        const startupElements = host.getElementsByClassName('theia-preload');
        return startupElements.length === 0 ? undefined : startupElements[0] as HTMLElement;
    }

    /* vvv HOTFIX begin vvv
     *
     * This is a hotfix against issues eclipse/theia#6459 and gitpod-io/gitpod#875 .
     * It should be reverted after Theia was updated to the newer Monaco.
     */
    protected inComposition = false;
    /**
     * Register composition related event listeners.
     */
    protected registerComositionEventListeners(): void {
        window.document.addEventListener('compositionstart', event => {
            this.inComposition = true;
        });
        window.document.addEventListener('compositionend', event => {
            this.inComposition = false;
        });
    }
    /* ^^^ HOTFIX end ^^^ */

    /**
     * Register global event listeners.
     */
    protected registerEventListeners(): void {
        this.registerComositionEventListeners(); /* Hotfix. See above. */

        window.addEventListener('beforeunload', () => {
            this.stateService.state = 'closing_window';
            // tslint:disable-next-line: no-any
            this.layoutRestorer.storeLayout(<any>this);
            this.stopContributions();
        });
        window.addEventListener('resize', () => this.shell.update());
        document.addEventListener('keydown', event => {
            if (this.inComposition !== true) {
                this.keybindings.run(event);
            }
        }, true);
        document.addEventListener('touchmove', event => { event.preventDefault(); }, { passive: false });
        // Prevent forward/back navigation by scrolling in OS X
        if (isOSX) {
            document.body.addEventListener('wheel', preventNavigation, { passive: false });
        }
        // Prevent the default browser behavior when dragging and dropping files into the window.
        window.addEventListener('dragover', event => {
            event.preventDefault();
        }, false);
        window.addEventListener('drop', event => {
            event.preventDefault();
        }, false);

    }

    /**
     * Attach the application shell to the host element. If a startup indicator is present, the shell is
     * inserted before that indicator so it is not visible yet.
     */
    protected attachShell(host: HTMLElement): void {
        const ref = this.getStartupIndicator(host);
        Widget.attach(this.shell, host, ref);
    }

    /**
     * If a startup indicator is present, it is first hidden with the `theia-hidden` CSS class and then
     * removed after a while. The delay until removal is taken from the CSS transition duration.
     */
    public revealShell(host: HTMLElement): Promise<void> {
        const startupElem = this.getStartupIndicator(host);
        if (startupElem) {
            return new Promise(resolve => {
                window.requestAnimationFrame(() => {
                    startupElem.classList.add('theia-hidden');
                    const preloadStyle = window.getComputedStyle(startupElem);
                    const transitionDuration = parseCssTime(preloadStyle.transitionDuration, 0);
                    window.setTimeout(() => {
                        const parent = startupElem.parentElement;
                        if (parent) {
                            parent.removeChild(startupElem);
                        }
                        resolve();
                    }, transitionDuration);
                });
            });
        } else {
            return Promise.resolve();
        }
    }

    /**
     * Initialize the shell layout either using the layout restorer service or, if no layout has
     * been stored, by creating the default layout.
     */
    protected async initializeLayout(): Promise<void> {
        if (!await this.restoreLayout()) {
            // Fallback: Create the default shell layout
            await this.createDefaultLayout();
        }
        await this.shell.pendingUpdates;
    }

    /**
     * Try to restore the shell layout from the storage service. Resolves to `true` if successful.
     */
    protected async restoreLayout(): Promise<boolean> {
        try {
            // tslint:disable-next-line: no-any
            return await this.layoutRestorer.restoreLayout(<any>this);
        } catch (error) {
            if (ApplicationShellLayoutMigrationError.is(error)) {
                console.warn(error.message);
                console.info('Initializing the default layout instead...');
            } else {
                console.error('Could not restore layout', error);
            }
            return false;
        }
    }

    /**
     * Let the frontend application contributions initialize the shell layout. Override this
     * method in order to create an application-specific custom layout.
     */
    protected async createDefaultLayout(): Promise<void> {
        for (const contribution of this.contributions.getContributions()) {
            if (contribution.initializeLayout) {
                await this.measure(contribution.constructor.name + '.initializeLayout',
                    // tslint:disable-next-line: no-any
                    () => contribution.initializeLayout!(<any>this)
                );
            }
        }
    }

    protected async fireOnDidInitializeLayout(): Promise<void> {
        for (const contribution of this.contributions.getContributions()) {
            if (contribution.onDidInitializeLayout) {
                await this.measure(contribution.constructor.name + '.onDidInitializeLayout',
                    // tslint:disable-next-line: no-any
                    () => contribution.onDidInitializeLayout!(<any>this)
                );
            }
        }
    }

    /**
     * Initialize and start the frontend application contributions.
     */
    protected async startContributions(): Promise<void> {
        for (const contribution of this.contributions.getContributions()) {
            if (contribution.initialize) {
                try {
                    contribution.initialize();
                } catch (error) {
                    this.logger.error('Could not initialize contribution', error);
                }
            }
        }

        for (const contribution of this.contributions.getContributions()) {
            if (contribution.configure) {
                try {
                    // tslint:disable-next-line: no-any
                    await contribution.configure(<any>this);
                } catch (error) {
                    this.logger.error('Could not configure contribution', error);
                }
            }
        }

        /**
         * FIXME:
         * - decouple commands & menus
         * - consider treat commands, keybindings and menus as frontend application contributions
         */
        this.commands.onStart();
        await this.keybindings.onStart();
        this.menus.onStart();
        for (const contribution of this.contributions.getContributions()) {
            if (contribution.onStart) {
                try {
                    await this.measure(contribution.constructor.name + '.onStart',
                        // tslint:disable-next-line: no-any
                        () => contribution.onStart!(<any>this)
                    );
                } catch (error) {
                    this.logger.error('Could not start contribution', error);
                }
            }
        }
    }

    /**
     * Stop the frontend application contributions. This is called when the window is unloaded.
     */
    protected stopContributions(): void {
        this.logger.info('>>> Stopping contributions....');
        for (const contribution of this.contributions.getContributions()) {
            if (contribution.onStop) {
                try {
                    // tslint:disable-next-line: no-any
                    contribution.onStop(<any>this);
                } catch (error) {
                    this.logger.error('Could not stop contribution', error);
                }
            }
        }
        this.logger.info('<<< All contributions have been stopped.');
    }

    protected async measure<T>(name: string, fn: () => MaybePromise<T>): Promise<T> {
        const startMark = name + '-start';
        const endMark = name + '-end';
        performance.mark(startMark);
        const result = await fn();
        performance.mark(endMark);
        performance.measure(name, startMark, endMark);
        for (const item of performance.getEntriesByName(name)) {
            if (item.duration > 100) {
                console.warn(item.name + ' is slow, took: ' + item.duration + ' ms');
            } else {
                console.debug(item.name + ' took ' + item.duration + ' ms');
            }
        }
        performance.clearMeasures(name);
        return result;
    }

}
