/********************************************************************************
 * Copyright (C) 2017 TypeFox and others.
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * This Source Code may also be made available under the following Secondary
 * Licenses when the conditions for such availability set forth in the Eclipse
 * Public License v. 2.0 are satisfied: GNU General Public License, version 2
 * with the GNU Classpath Exception which is available at
 * https://www.gnu.org/software/classpath/license.html.
 *
 * SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
 ********************************************************************************/

import * as os from 'os';
import { injectable, inject, optional } from 'inversify';
import { TextDocumentContentChangeEvent, Range } from 'vscode-languageserver-protocol';
import { FileUri } from '@theia/core/lib/node/file-uri';
import { FileStat, FileSystem, FileSystemClient, FileSystemError, FileMoveOptions, FileDeleteOptions, FileAccess } from '../common/filesystem';
import { RemoteHostsManager, RemoteFileSystem } from '@theia/ct-core/lib/common';

@injectable()
export class FileSystemNodeOptions {

    encoding: string;
    recursive: boolean;
    overwrite: boolean;
    moveToTrash: boolean;

    public static DEFAULT: FileSystemNodeOptions = {
        encoding: 'utf8',
        overwrite: false,
        recursive: true,
        moveToTrash: true
    };

}

@injectable()
export class FileSystemRemote implements FileSystem {
    // codetogether-start
    public hostManager: RemoteHostsManager;

    // codetogether-end
    constructor(
        @inject(FileSystemNodeOptions) @optional() protected readonly options: FileSystemNodeOptions = FileSystemNodeOptions.DEFAULT
    ) { }

    protected client: FileSystemClient | undefined;
    setClient(client: FileSystemClient | undefined): void {
        this.client = client;
    }

    async getFileStat(uri: string): Promise<FileStat | undefined> {
        const remoteFileSystem = this.getRemoteFileSystemFor(uri);
        if (remoteFileSystem) {
            return remoteFileSystem.getFileStat(uri);
        }
        throw FileSystemError.NoHostForFile(uri);
    }

    async exists(uri: string): Promise<boolean> {
        const remoteFileSystem = this.getRemoteFileSystemFor(uri);
        if (remoteFileSystem) {
            return remoteFileSystem.exists(uri);
        }
        throw FileSystemError.NoHostForFile(uri);
    }

    async resolveContent(uri: string, options?: { encoding?: string }): Promise<{
        stat: FileStat, content: string, dirty: boolean,
        revision: { origin: string, seq: number }, selection?: Range,
        encrypted?: boolean, compressed?: boolean
    }> { // codetegether
        const remoteFileSystem = this.getRemoteFileSystemFor(uri);
        if (remoteFileSystem) {
            return remoteFileSystem.resolveContent(uri);
        }
        throw FileSystemError.NoHostForFile(uri);
    }

    async setContent(file: FileStat, content: string, options?: { encoding?: string, encrypted?: boolean, compressed?: boolean }): Promise<FileStat> {
        const remoteFileSystem = this.getRemoteFileSystemFor(file.uri);
        if (remoteFileSystem) {
            return remoteFileSystem.setContent(file, content, options);
        }
        throw FileSystemError.NoHostForFile(file.uri);
    }

    async updateContent(file: FileStat, contentChanges: TextDocumentContentChangeEvent[], options?: { encoding?: string, overwriteEncoding?: string }): Promise<FileStat> {
        const remoteFileSystem = this.getRemoteFileSystemFor(file.uri);
        if (remoteFileSystem) {
            return remoteFileSystem.updateContent(file, contentChanges, options);
        }
        throw FileSystemError.NoHostForFile(file.uri);
    }

    async move(sourceUri: string, targetUri: string, options?: FileMoveOptions): Promise<FileStat> {
        const remoteFileSystem = this.getRemoteFileSystemFor(sourceUri);
        if (remoteFileSystem) {
            return remoteFileSystem.move(sourceUri, targetUri, options);
        }
        throw FileSystemError.NoHostForFile(sourceUri);
    }

    async copy(sourceUri: string, targetUri: string, options?: { overwrite?: boolean, recursive?: boolean }): Promise<FileStat> {
        const remoteFileSystem = this.getRemoteFileSystemFor(sourceUri);
        if (remoteFileSystem) {
            return remoteFileSystem.copy(sourceUri, targetUri, options);
        }
        throw FileSystemError.NoHostForFile(sourceUri);
    }

    async createFile(uri: string, options?: { content?: string, encoding?: string, encrypted?: boolean, compressed?: boolean }): Promise<FileStat> {
        const remoteFileSystem = this.getRemoteFileSystemFor(uri);
        if (remoteFileSystem) {
            return remoteFileSystem.createFile(uri, options);
        }
        throw FileSystemError.NoHostForFile(uri);
    }

    async createFolder(uri: string): Promise<FileStat> {
        const remoteFileSystem = this.getRemoteFileSystemFor(uri);
        if (remoteFileSystem) {
            return remoteFileSystem.createFolder(uri);
        }
        throw FileSystemError.NoHostForFile(uri);
    }

    async touchFile(uri: string): Promise<FileStat> {
        const remoteFileSystem = this.getRemoteFileSystemFor(uri);
        if (remoteFileSystem) {
            return remoteFileSystem.touchFile(uri);
        }
        throw FileSystemError.NoHostForFile(uri);
    }

    async delete(uri: string, options?: FileDeleteOptions): Promise<void> {
        const remoteFileSystem = this.getRemoteFileSystemFor(uri);
        if (remoteFileSystem) {
            return remoteFileSystem.delete(uri, options);
        }
        throw FileSystemError.NoHostForFile(uri);
    }

    async getEncoding(uri: string): Promise<string> {
        const remoteFileSystem = this.getRemoteFileSystemFor(uri);
        if (remoteFileSystem) {
            return remoteFileSystem.getEncoding(uri);
        }
        throw FileSystemError.NoHostForFile(uri);
    }

    async guessEncoding(uri: string): Promise<string | undefined> {
        const remoteFileSystem = this.getRemoteFileSystemFor(uri);
        if (remoteFileSystem) {
            return remoteFileSystem.guessEncoding(uri);
        }
        throw FileSystemError.NoHostForFile(uri);
    }

    async getRoots(): Promise<FileStat[]> {
        if (this.hostManager) {
            const hosts = new Array<Promise<FileStat[]>>();
            this.hostManager.getRemoteHosts().forEach(host => {
                hosts.push(host.getRemoteFileSystem().getRoots());
            });
            return Promise.all(hosts).then(resolved => {
                const roots = new Array<FileStat>();
                resolved.forEach(next => {
                    next.forEach(n => roots.push(n));
                });
                return roots;
            });
        }
        return [];
    }

    async getCurrentUserHome(): Promise<FileStat | undefined> {
        return this.getFileStat(FileUri.create(os.homedir()).toString());
    }

    getDrives(): Promise<string[]> {
        throw FileSystemError.NoHostForFile();
    }

    dispose(): void {
        // NOOP
    }

    async access(uri: string, mode: number = FileAccess.Constants.F_OK): Promise<boolean> {
        const remoteFileSystem = this.getRemoteFileSystemFor(uri);
        if (remoteFileSystem) {
            return remoteFileSystem.access(uri, mode);
        }
        throw FileSystemError.NoHostForFile(uri);
    }

    async getFsPath(uri: string): Promise<string | undefined> {
        const remoteFileSystem = this.getRemoteFileSystemFor(uri);
        if (remoteFileSystem) {
            return remoteFileSystem.getFsPath(uri);
        }
        throw FileSystemError.NoHostForFile(uri);
    }

    private getRemoteFileSystemFor(uri: string): RemoteFileSystem | undefined {
        if (this.hostManager) {
            const host = this.hostManager.getRemoteHostFor(uri);
            if (host) {
                return host.getRemoteFileSystem();
            }
        }
        return undefined;
    }
}
