/********************************************************************************
 * Copyright (C) 2017 TypeFox and others.
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * This Source Code may also be made available under the following Secondary
 * Licenses when the conditions for such availability set forth in the Eclipse
 * Public License v. 2.0 are satisfied: GNU General Public License, version 2
 * with the GNU Classpath Exception which is available at
 * https://www.gnu.org/software/classpath/license.html.
 *
 * SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
 ********************************************************************************/

import { injectable, inject } from 'inversify';
import { MessageConnection } from 'vscode-jsonrpc';
import { CommandRegistry, Disposable, MaybePromise } from '@theia/core/lib/common';
import {
    ErrorAction, RevealOutputChannelOn, CloseAction, IConnection, DidChangeTextDocumentParams,
    DidOpenTextDocumentParams, NotificationHandler, ShowMessageParams, LogMessageParams, DidChangeConfigurationParams,
    DidChangeWatchedFilesParams, DidCloseTextDocumentParams, DidSaveTextDocumentParams, PublishDiagnosticsParams,
    InitializeParams, InitializeResult, Trace, Tracer
} from 'monaco-languageclient';
import {
    Workspace, Languages, Window, Services,
    ILanguageClient, LanguageClientOptions, MonacoLanguageClient,
    createConnection, LanguageContribution
} from './language-client-services';
import { TypeHierarchyFeature } from './typehierarchy/typehierarchy-feature';
import { EndToEndService, EndToEndServiceImpl } from '@theia/core/lib/browser/codetogether-startup/session';

@injectable()
export class LanguageClientFactory {

    @inject(CommandRegistry) protected readonly registry: CommandRegistry;

    constructor(
        @inject(Workspace) protected readonly workspace: Workspace,
        @inject(Languages) protected readonly languages: Languages,
        @inject(Window) protected readonly window: Window,
        @inject(EndToEndService) protected readonly endToEnd: EndToEndServiceImpl
    ) {
        Services.install({
            workspace,
            languages,
            window,
            commands: {
                registerCommand: this.registerCommand.bind(this)
            }
        });
    }

    // tslint:disable-next-line:no-any
    protected registerCommand(id: string, callback: (...args: any[]) => any, thisArg?: any): Disposable {
        const execute = callback.bind(thisArg);
        return this.registry.registerCommand({ id }, { execute });
    }

    get(contribution: LanguageContribution, clientOptions: LanguageClientOptions,
        connectionProvider: MessageConnection | (() => MaybePromise<MessageConnection>)): ILanguageClient {
        if (clientOptions.revealOutputChannelOn === undefined) {
            clientOptions.revealOutputChannelOn = RevealOutputChannelOn.Never;
        }
        if (!clientOptions.errorHandler) {
            clientOptions.errorHandler = {
                // ignore connection errors
                error: () => ErrorAction.Continue,
                closed: () => CloseAction.DoNotRestart
            };
        }
        const initializationFailedHandler = clientOptions.initializationFailedHandler;
        clientOptions.initializationFailedHandler = e => !!initializationFailedHandler && initializationFailedHandler(e);
        const client = new MonacoLanguageClient({
            id: contribution.id,
            name: contribution.name,
            clientOptions,
            connectionProvider: {
                get: async (errorHandler, closeHandler) => {
                    const connection = typeof connectionProvider === 'function' ? await connectionProvider() : connectionProvider;
                    return new EncryptingConnection(createConnection(connection, errorHandler, closeHandler), this.endToEnd);
                }
            }
        });
        client.registerFeature(new TypeHierarchyFeature(client));
        return this.patch4085(client);
    }

    /**
     * see https://github.com/eclipse-theia/theia/issues/4085
     * remove when monaco-languageclient is upgraded to latest vscode-languageclient
     */
    protected patch4085(client: MonacoLanguageClient): MonacoLanguageClient {
        const features = client['_dynamicFeatures'] as Map<string, {
            _listener?: Object | undefined
            dispose?: Function
        }>;
        for (const feature of features.values()) {
            if (feature.dispose) {
                const dispose = feature.dispose.bind(feature);
                feature.dispose = () => {
                    dispose();
                    if (feature._listener) {
                        feature._listener = undefined;
                    }
                };
            }
        }
        return client;
    }

}

// codetogether-class
class EncryptingConnection implements IConnection {

    constructor(private connection: IConnection, private endToEnd: EndToEndServiceImpl) {
    }

    // tslint:disable-next-line:no-any
    async sendRequest(method: any, param?: any, token?: any): Promise<any> {
        if (this.endToEnd.isEndToEnd() && param.textDocument) {
            param = [this.endToEnd.session, await this.endToEnd.encrypt(JSON.stringify(param))];
        } else if (this.endToEnd.isEndToEnd() && param.command && param.arguments) {
            param.session = this.endToEnd.session;
            param.arguments = [await this.endToEnd.encrypt(JSON.stringify(param.arguments))];
        } else {
            // this was historically not routed on backend side
            // now allowing it to proceed using code below causes errors
            // on host side as this is not used
            return Promise.resolve();
        }
        return this.connection.sendRequest(method, param, token).then(result => {
            if (this.endToEnd.isEndToEnd()) {
                if (result && typeof result === 'string') {
                    return this.endToEnd.decrypt(result, false).then(decoded => JSON.parse(decoded));
                }
            }
            return result;
        });
    }

    // tslint:disable-next-line:no-any
    onRequest(method: any, handler: any): void {
        this.connection.onRequest(method, handler);
    }

    listen(): void {
        this.connection.listen();
    }

    // tslint:disable-next-line:no-any
    sendNotification(method: any, params?: any): void {
        this.connection.sendNotification(method, params);
    }

    // tslint:disable-next-line:no-any
    onNotification(method: any, handler: any): void {
        this.connection.onNotification(method, handler);
    }
    trace(value: Trace, tracer: Tracer, sendNotification?: boolean | undefined): void {
        this.connection.trace(value, tracer, sendNotification);
    }
    // tslint:disable-next-line:no-any
    initialize(params: InitializeParams): Thenable<InitializeResult<any>> {
        return this.connection.initialize(params);
    }
    shutdown(): Thenable<void> {
        return this.connection.shutdown();
    }
    exit(): void {
        this.connection.exit();
    }
    onLogMessage(handle: NotificationHandler<LogMessageParams>): void {
        this.connection.onLogMessage(handle);
    }
    onShowMessage(handler: NotificationHandler<ShowMessageParams>): void {
        this.connection.onShowMessage(handler);
    }

    // tslint:disable-next-line:no-any
    onTelemetry(handler: NotificationHandler<any>): void {
        this.connection.onTelemetry(handler);
    }
    didChangeConfiguration(params: DidChangeConfigurationParams): void {
        this.connection.didChangeConfiguration(params);
    }
    didChangeWatchedFiles(params: DidChangeWatchedFilesParams): void {
        this.connection.didChangeWatchedFiles(params);
    }
    didOpenTextDocument(params: DidOpenTextDocumentParams): void {
        this.connection.didOpenTextDocument(params);
    }
    didChangeTextDocument(params: DidChangeTextDocumentParams): void {
        this.connection.didChangeTextDocument(params);
    }
    didCloseTextDocument(params: DidCloseTextDocumentParams): void {
        this.connection.didCloseTextDocument(params);
    }
    didSaveTextDocument(params: DidSaveTextDocumentParams): void {
        this.connection.didSaveTextDocument(params);
    }
    onDiagnostics(handler: NotificationHandler<PublishDiagnosticsParams>): void {
        this.connection.onDiagnostics(handler);
    }
    dispose(): void {
        this.connection.dispose();
    }
}
